const char uri_lua[] =
"-- uri.lua (internal file)\n"
"\n"
"local ffi = require('ffi')\n"
"local buffer = require('buffer')\n"
"local uri = require('uri')\n"
"\n"
"ffi.cdef[[\n"
"struct uri_param {\n"
"    const char *name;\n"
"    int value_count;\n"
"    const char **values;\n"
"};\n"
"\n"
"/**\n"
" * We define all strings inside the `struct uri` as const, despite\n"
" * the fact that they are not constant in the C structure. This is\n"
" * necessary for the `uri_format` function work: there we cannot\n"
" * assign lua strings to a non-constant pointer. It's not a problem\n"
" * since uri_format doesn't change something in `struct uri`.\n"
" */\n"
"struct uri {\n"
"    const char *scheme;\n"
"    const char *login;\n"
"    const char *password;\n"
"    const char *host;\n"
"    const char *service;\n"
"    const char *path;\n"
"    const char *query;\n"
"    const char *fragment;\n"
"    int host_hint;\n"
"    int param_count;\n"
"    struct uri_param *params;\n"
"};\n"
"\n"
"struct uri_set {\n"
"    int uri_count;\n"
"    struct uri *uris;\n"
"};\n"
"\n"
"void\n"
"uri_destroy(struct uri *uri);\n"
"\n"
"void\n"
"uri_set_destroy(struct uri_set *uri_set);\n"
"\n"
"int\n"
"uri_format(char *str, size_t len, struct uri *uri, bool write_password);\n"
"]]\n"
"\n"
"local builtin = ffi.C;\n"
"local uri_stash = buffer.ffi_stash_new('struct uri')\n"
"local uri_stash_take = uri_stash.take\n"
"local uri_stash_put = uri_stash.put\n"
"local cord_ibuf_take = buffer.internal.cord_ibuf_take\n"
"local cord_ibuf_put = buffer.internal.cord_ibuf_put\n"
"\n"
"local uri_set_stash = buffer.ffi_stash_new('struct uri_set')\n"
"local uri_set_stash_take = uri_set_stash.take\n"
"local uri_set_stash_put = uri_set_stash.put\n"
"\n"
"local function parse_uribuf(uribuf)\n"
"    local result = {}\n"
"    for _, k in ipairs({ 'scheme', 'login', 'password', 'host', 'service',\n"
"        'path', 'query', 'fragment'}) do\n"
"        if uribuf[k] ~= nil then\n"
"            result[k] = ffi.string(uribuf[k])\n"
"        end\n"
"    end\n"
"    if uribuf.param_count > 0 then\n"
"        result.params = {}\n"
"        for param_idx = 0, uribuf.param_count - 1 do\n"
"            local param = uribuf.params[param_idx]\n"
"            local name = ffi.string(param.name)\n"
"            result.params[name] = {}\n"
"            for val_idx = 0, param.value_count - 1 do\n"
"                result.params[name][val_idx + 1] =\n"
"                    ffi.string(param.values[val_idx])\n"
"            end\n"
"        end\n"
"    end\n"
"    if uribuf.host_hint == 1 then\n"
"        result.ipv4 = result.host\n"
"    elseif uribuf.host_hint == 2 then\n"
"        result.ipv6 = result.host\n"
"    elseif uribuf.host_hint == 3 then\n"
"        result.unix = result.service\n"
"    end\n"
"    return result\n"
"end\n"
"\n"
"local function parse(str)\n"
"    if str == nil then\n"
"        error(\"Usage: uri.parse(string|table)\")\n"
"    end\n"
"    local uribuf = uri_stash_take()\n"
"    local status, errmsg = pcall(uri.internal.uri_create, uribuf, str)\n"
"    if not status then\n"
"        uri_stash_put(uribuf)\n"
"        return nil, errmsg\n"
"    end\n"
"    local result = parse_uribuf(uribuf)\n"
"    builtin.uri_destroy(uribuf)\n"
"    uri_stash_put(uribuf)\n"
"    return result\n"
"end\n"
"\n"
"local function parse_many(str)\n"
"    if str == nil then\n"
"        error(\"Usage: uri.parse_many(string|table)\")\n"
"    end\n"
"    local uri_set_buf = uri_set_stash_take()\n"
"    local status, errmsg = pcall(uri.internal.uri_set_create, uri_set_buf, str)\n"
"    if not status then\n"
"        uri_set_stash_put(uri_set_buf)\n"
"        return nil, errmsg\n"
"    end\n"
"    local result = {}\n"
"    for i = 0, uri_set_buf.uri_count - 1 do\n"
"        result[i + 1] = parse_uribuf(uri_set_buf.uris[i])\n"
"    end\n"
"    builtin.uri_set_destroy(uri_set_buf)\n"
"    uri_set_stash_put(uri_set_buf)\n"
"    return result\n"
"end\n"
"\n"
"local function fill_uribuf_params(uribuf, uri)\n"
"    uribuf.param_count = 0\n"
"    uribuf.params = nil\n"
"    if not uri.params then\n"
"        return\n"
"    end\n"
"    for _, _ in pairs(uri.params) do\n"
"        uribuf.param_count = uribuf.param_count + 1\n"
"    end\n"
"    if uribuf.param_count == 0 then\n"
"        return\n"
"    end\n"
"    uribuf.params = ffi.new(\"struct uri_param[\?]\", uribuf.param_count)\n"
"    local i = 0\n"
"    for param_name, param in pairs(uri.params) do\n"
"        uribuf.params[i].value_count = #param\n"
"        uribuf.params[i].name = param_name\n"
"        uribuf.params[i].values =\n"
"            ffi.new(\"const char *[\?]\", uribuf.params[i].value_count)\n"
"        for j = 1, uribuf.params[i].value_count do\n"
"            uribuf.params[i].values[j - 1] = param[j]\n"
"        end\n"
"        i = i + 1\n"
"    end\n"
"end\n"
"\n"
"local function format(uri, write_password)\n"
"    local uribuf = uri_stash_take()\n"
"    uribuf.scheme = uri.scheme\n"
"    uribuf.login = uri.login\n"
"    uribuf.password = uri.password\n"
"    uribuf.host = uri.host\n"
"    uribuf.service = uri.service\n"
"    uribuf.path = uri.path\n"
"    uribuf.query = uri.query\n"
"    uribuf.fragment = uri.fragment\n"
"    fill_uribuf_params(uribuf, uri)\n"
"    local ibuf = cord_ibuf_take()\n"
"    local str = ibuf:alloc(1024)\n"
"    local len = builtin.uri_format(str, 1024, uribuf, write_password and 1 or 0)\n"
"    uri_stash_put(uribuf)\n"
"    str = ffi.string(str, len)\n"
"    cord_ibuf_put(ibuf)\n"
"    return str\n"
"end\n"
"\n"
"return {\n"
"    parse_many = parse_many,\n"
"    parse = parse,\n"
"    format = format,\n"
"};\n"
""
;
